/*
 *	Calendar program - one month per page
 *
 *	Originally written in FORTRAN-IV for GE Timesharing, 10/65
 *	Re-coded in C for UNIX, 3/83
 *
 *	Author: AW Rogers
 *
 *	Parameters:
 *
 *		calen yy		generates calendar for year yy
 *
 *		calen mm yy [len]	generates calendar for len months
 *					(default = 1) starting with mm/yy
 *
 * 	Option flags (must precede params):
 *
 *		-l			left-justify dates (default)
 *		-r			right-justify dates
 *		-m			mixed-case output (default)
 *		-u			upper-case output
 *		-o[seq]			use "seq" as overstrike sequence
 *					 for heading (default: HIX)
 *		-bN			add N blank lines after form-feed
 *
 *		Output is to standard output.
 *		
 */
 
#include <stdio.h>
#include <ctype.h>

#define FALSE   0
#define TRUE    1
 
#define JAN 	1			/* significant months/years */
#define FEB	2
#define DEC	12
#define MINYR	1753
#define MAXYR	9999
 
#define SOLID	0			/* pseudo-enumeration for line styles */
#define OPEN	1

#define LEFT	0			/* ... and justification of dates     */
#define RIGHT	1

#define MIXED	0			/* ... and case of output text	      */

#define UPPER	1

#define	OVERSTRIKE	"HIX"		/* overstrike sequence for month/year */
#define MAX_OVERSTR	3		/* maximum overstrikes permitted */

#define isLeap(y) ((y) % 4 == 0 && ((y) % 100 != 0 || (y) % 400 == 0))	/* leap year macro */
 
typedef struct				/* info for a single month */
    {
    int 	mm;
    int		yy;
    char 	mmname[10];
    char	dates[6][7][3];
    } monthRec;
 
typedef monthRec *mptr;			/* pointer to above struct */

/* globals corresponding to command-line flags */

int just = LEFT;			/* default justification of dates */
int ocase = MIXED;			/* default case for output */
int nblank = 0;				/* default blank lines after FF */
char *seq = OVERSTRIKE;			/* default overstrike sequence */


/*
 *	Main - gets and validates parameters, opens output file, executes
 *	loop to fill and print months of calendar, closes output file
 */
main(argc, argv)
    int argc;
    char *argv[];
    {
    int nmonths;			/* consecutive months to print */
    int badopt = FALSE;			/* flag set if bad option */
    int badpar = FALSE;			/* flag set if bad param  */
    monthRec mRec[3];			/* space for main and small calendars */
    mptr prev = &mRec[0], 		/* pointers to calendars (initially) */
	 curr = &mRec[1],
	 next = &mRec[2],
	 temp;

    /* Get command line flags */

    while (argc > 1 && argv[1][0] == '-')
        {
        switch (argv[1][1])
	    {
	    case 'b':
		sscanf(&argv[1][2], "%d", &nblank);
		break;
	    case 'l':
		just = LEFT;
		break;
	    case 'r':
		just = RIGHT;
		break;
	    case 'm':
		ocase = MIXED;
		break;
	    case 'u':
		ocase = UPPER;
		break;
	    case 'o':
		if (argv[1][2] != '\0')
		    seq = &argv[1][2];
		break;
	    default:
		fprintf(stderr, "Invalid flag: %s\n", argv[1]);
		badopt = TRUE;
		break;
	    }
        argv++;
	argc--;
	}

    if (badopt)
	fprintf(stderr, "Valid flags are -b -l -m -o -r -u\n");

    /* Get and validate parameters */
    
    if (argc == 2)		/* only one arg - treat as yy */
	{
	sscanf(argv[1], "%d", &curr->yy);
	curr->mm = JAN;
	nmonths = 12;
	}
 
    else if (argc >= 3)		/* two or more - treat as mm yy [len] */
	{
	sscanf(argv[1], "%d", &curr->mm);
	sscanf(argv[2], "%d", &curr->yy);
	if (argc >= 4)
	    sscanf(argv[3], "%d", &nmonths);
	}
 
    else			/* none specified - get interactively */
	{
	fprintf(stderr, "Enter calendar specs (month year length): ");
	scanf("%d %d %d", &curr->mm, &curr->yy, &nmonths);
	}
 
    if (curr->yy > 0 && curr->yy < 100)		/* nn -> 19nn */
	curr->yy += 1900;
 
    if (nmonths < 1)				/* default for month count */
	nmonths = 1;
 
    if (curr->mm < JAN || curr->mm > DEC)	/* validate month/year */
      {
      fprintf(stderr, "Month %d not in range %d .. %d\n", curr->mm, JAN, DEC);
      badpar = TRUE;
      }

    if (curr->yy < MINYR || curr->yy > MAXYR)
      {
      fprintf(stderr, "Year %d not in range %d .. %d\n", curr->yy, MINYR,
	      MAXYR);
      badpar = TRUE;
      }

    if (badpar)		/* quit if month or year invalid */
      exit(1);
 
    /* fill in calendars for previous and current month */
 
    prev->mm = (curr->mm == JAN) ? DEC : curr->mm - 1;
    prev->yy = (curr->mm == JAN) ? curr->yy - 1 : curr->yy;
    fillCalendar(prev);
 
    fillCalendar(curr);
 
    /*
     * Main loop: print each month of the calendar (with small calendars for
     * the preceding and following months in the upper corners).  The current
     * and next months' calendars can be reused the following month; only
     * the 'next' calendar need be recalculated each time.
     */
 
    for (; nmonths > 0 && curr->yy <= MAXYR; nmonths--)	/* main loop */
	{
	next->mm = (curr->mm == DEC) ? JAN : curr->mm + 1;
	next->yy = (curr->mm == DEC) ? curr->yy + 1 : curr->yy;
	fillCalendar(next);			/* fill in following month */
 
	printCalendar(prev, curr, next);
 
	temp = prev;				/* swap the three months */
	prev = curr;
	curr = next;
	next = temp;
	}
 
    }

/*
 *	Print the calendar for the current month, generating small calendars
 *	for the previous and following months in the upper corners and the
 *	month/year (in large characters) centered at the top.
 */
printCalendar(prev, curr, next)
    mptr prev;			/* Previous month (upper-left corner)  */
    mptr curr;			/* Current month (main calendar)       */
    mptr next;			/* Next month (upper-right corner)     */
    {
    int nchars, i, j;
    static char *mc_wkday[] =
	{
	" Sunday  ", " Monday  ", " Tuesday ", "Wednesday", "Thursday ",
	" Friday  ", "Saturday "
	};
    static char *uc_wkday[] =
	{
	" SUNDAY  ", " MONDAY  ", " TUESDAY ", "WEDNESDAY", "THURSDAY ",
	" FRIDAY  ", "SATURDAY "
	};

    char **wkday;				/* pointer to one of above */
    char *blanks = "                     "; 	/* 21 blanks for centering */
    char *padding;				/* Pointer into 'blanks'   */
    char monthAndYear[20];			/* Work area 		   */
    char *ovr;					/* overstrike sequence	   */
 
    nchars = strlen(curr->mmname);		/* set up month/year heading */
    padding = blanks + (3 * (nchars - 3));	/* and center it  	     */
    sprintf(monthAndYear, "%s%5d", curr->mmname, curr->yy);
 
    printf("\f\n");			/* print month/year in large chars */
    for (i = 0; i < nblank; i++)
	printf("\n");

    for (i = 0; i < 9; i++)		/* surrounded by small calendars   */
        {
	for (ovr = seq;			/* overstruck lines first */
	     ovr < seq + MAX_OVERSTR - 1 && *(ovr+1);
	     ovr++) 
	    {
	    printf("%20s%s", " ", padding);
	    printHdr(monthAndYear, i, *ovr);
	    printf("\r");
	    }
	printSmallCal(prev, i);		/* then small calendars, etc. */
	printf("%s", padding);
	printHdr(monthAndYear, i, *ovr);
	printf(" %s", padding);
        printSmallCal(next, i);
        printf("\n");
        }
 
    printf("\n");			/* print the weekday names */
    print_line(1, SOLID);
    print_line(1, OPEN);
    printf("  ");
    wkday = ocase == UPPER ? uc_wkday : mc_wkday;
    for (j = 0; j < 7; j++)
        printf("|%13.9s    ", wkday[j]);
    printf("|\n");
    print_line(1, OPEN);
    
    for (i = 0; i < 4; i++)		/* print first four rows */
        {
        print_line(1, SOLID);
        print_dates(curr, i, just);
        print_line(7, OPEN);
        }
 
    print_line(1, SOLID);		/* print bottom row */
    print_dates(curr, 4, just);
    print_line(2, OPEN);
    print_divider(curr->dates[5]);	/* divider for 23/30, 24/31 */

    print_line(3, OPEN);
    print_dates(curr, 5, !just);	/* print 6th line (30/31) at bottom */
    print_line(1, SOLID);
    }

/*
 *	Fill in the month name and date fields of a specified calendar record
 *	(assumes mm, yy fields already filled in)
 */
fillCalendar(month)
    mptr month;			/* Pointer to month info record */
    {
    typedef struct		/* Local info about months */
	{
	char *name[2];		/* Name of month (mixed/upper-case)	*/
	int offset[2];		/* Offset of m/1 from 1/1 (non-leap/leap) */
	int length[2];		/* Length (non-leap/leap)		*/
	} monthInfo;
 
    static monthInfo info[12] = {
	{ {"January",	"JANUARY"},	{0, 0},	{31, 31} },
	{ {"February",	"FEBRUARY"},	{3, 3},	{28, 29} },
	{ {"March",	"MARCH"},	{3, 4},	{31, 31} },
	{ {"April",	"APRIL"},	{6, 0},	{30, 30} },
	{ {"May",	"MAY"},		{1, 2},	{31, 31} },
	{ {"June",	"JUNE"},	{4, 5},	{30, 30} },
	{ {"July",	"JULY"},	{6, 0},	{31, 31} },
	{ {"August",	"AUGUST"},	{2, 3},	{31, 31} },
	{ {"September",	"SEPTEMBER"},	{5, 6},	{30, 30} },
	{ {"October",	"OCTOBER"},	{0, 1},	{31, 31} },
	{ {"November",	"NOVEMBER"},	{3, 4},	{30, 30} },
	{ {"December",	"DECEMBER"},	{5, 6},	{31, 31} }
    };
 
    int i, first, last, date = 0, y = month->yy, m = month->mm-1;
    int leap = isLeap(y);
 
    first = (y + (y-1)/4 - (y-1)/100 + (y-1)/400 + info[m].offset[leap]) % 7;
    last = first + info[m].length[leap] - 1;
 
    for (i = 0; i < 42; i++)			/* fill in the dates */
	if (i < first || i > last)
	    strcpy(month->dates[i/7][i%7], "  ");
	else
	    sprintf(month->dates[i/7][i%7], "%2d", ++date);
 
    strcpy(month->mmname, info[m].name[ocase]);	/* copy name of month */
    } 

/*
 *	Print one line of a small calendar (previous and next months in
 *	upper left and right corners of output)
 */
printSmallCal(month, line)
    mptr month;			/* Month info record pointer	*/
    int line;			/* Line to print (see below)	*/
    {
    int i;
  
    switch (line)
	{
	case 0:		/* month/year at top */
	    printf("   %-10s%4d   ", month->mmname, month->yy);
	    break;
	case 1:		/* blank line */
	    printf("%20s", " ");
	    break;
	case 2:		/* weekdays */
	    printf(ocase == UPPER ? "SU MO TU WE TH FR SA" :
				    "Su Mo Tu We Th Fr Sa");
	    break;
	default:	/* line of calendar */
	    for (i = 0; i <= 5; i++)
    		printf("%s ", month->dates[line-3][i]);
	    printf("%s", month->dates[line-3][6]);
	    break;
	}
    }

/*
 *	Print n lines in selected style
 */
print_line(n, style)
    int n;			/* Number of lines to print (> 0)   */
    int style;			/* SOLID or OPEN 		    */
    {
    int i;
    char *fmt1 = (style == SOLID) ? "+-----------------" :
				    "|                 " ;
    char *fmt2 = (style == SOLID) ? "+\n" : "|\n" ;

    for (; n > 0; n--)
	{
	printf("  ");
	for (i = 0; i < 7; i++)
	    printf(fmt1);
	printf(fmt2);
	}
    }



/*
 *	Print line of large calendar (open w/left- or right-justified dates)
 */
print_dates(month, line, just)
    mptr month;			/* Month info record pointer	*/
    int line;			/* Line to print (0-5)		*/
    int just;			/* justification (LEFT / RIGHT)	*/
    {
    int i;
    char *fmt = (just == LEFT) ? "| %-16s" : "|%16s " ;

    printf("  ");
    for (i = 0; i < 7; i++)
	printf(fmt, month->dates[line][i]);
    printf("|\n");
    }

/*
 *	Print divider between 23/30 and 24/31
 */
print_divider(dates)
    char dates[7][3];
    {
    int j;

    printf("  ");
    for (j = 0; j < 7; j++)
        if (strcmp(dates[j], "  ") == 0)
    	    printf("|                 ");
        else
    	    printf("|_________________");
    printf("|\n");
    }

 
/*
 *	Print LS 6 bits of n (0 = ' '; 1 = selected non-blank)
 */
decode(n, c)
    int n;				/* Number to print (0-31) */
    char c;
    {
    int msk = 1 << 5;
 
    for (; msk; msk /= 2) 
	printf("%c", (n & msk) ? c : ' ');
    }
 
 
/*
 *	Print one line of string in large characters
 */
printHdr(str, line, c)
    char *str;				/* string to print		    */
    int line;				/* line to print (0-8; else blanks) */
    char c;				/* output character to use	    */
    {
 
    /* 5x9 dot-matrix representations of A-Z, a-z, 0-9 */
 
    static char uppers[26][9] =
	{
	{14, 17, 17, 31, 17, 17, 17,  0,  0},	{30, 17, 17, 30, 17, 17, 30,  0,  0},	/* AB */
	{14, 17, 16, 16, 16, 17, 14,  0,  0},	{30, 17, 17, 17, 17, 17, 30,  0,  0},	/* CD */
	{31, 16, 16, 30, 16, 16, 31,  0,  0},	{31, 16, 16, 30, 16, 16, 16,  0,  0},	/* EF */
	{14, 17, 16, 23, 17, 17, 14,  0,  0},	{17, 17, 17, 31, 17, 17, 17,  0,  0},	/* GH */
	{31,  4,  4,  4,  4,  4, 31,  0,  0},	{ 1,  1,  1,  1,  1, 17, 14,  0,  0},	/* IJ */
	{17, 18, 20, 24, 20, 18, 17,  0,  0},	{16, 16, 16, 16, 16, 16, 31,  0,  0},	/* KL */
	{17, 27, 21, 21, 17, 17, 17,  0,  0},	{17, 17, 25, 21, 19, 17, 17,  0,  0},	/* MN */
	{14, 17, 17, 17, 17, 17, 14,  0,  0},	{30, 17, 17, 30, 16, 16, 16,  0,  0},	/* OP */
	{14, 17, 17, 17, 21, 18, 13,  0,  0},	{30, 17, 17, 30, 20, 18, 17,  0,  0},	/* QR */
	{14, 17, 16, 14,  1, 17, 14,  0,  0},	{31,  4,  4,  4,  4,  4,  4,  0,  0},	/* ST */
	{17, 17, 17, 17, 17, 17, 14,  0,  0},	{17, 17, 17, 17, 17, 10,  4,  0,  0},	/* UV */
	{17, 17, 17, 21, 21, 21, 10,  0,  0},	{17, 17, 10,  4, 10, 17, 17,  0,  0},	/* WX */
	{17, 17, 17, 14,  4,  4,  4,  0,  0},	{31,  1,  2,  4,  8, 16, 31,  0,  0}	/* YZ */
	}; 
 
    static char lowers[26][9] =
	{
	{ 0,  0, 14,  1, 15, 17, 15,  0,  0},	{16, 16, 30, 17, 17, 17, 30,  0,  0},	/* ab */
	{ 0,  0, 15, 16, 16, 16, 15,  0,  0},	{ 1,  1, 15, 17, 17, 17, 15,  0,  0},	/* cd */
	{ 0,  0, 14, 17, 31, 16, 14,  0,  0},	{ 6,  9, 28,  8,  8,  8,  8,  0,  0},	/* ef */
	{ 0,  0, 14, 17, 17, 17, 15,  1, 14},	{16, 16, 30, 17, 17, 17, 17,  0,  0},	/* gh */
	{ 4,  0, 12,  4,  4,  4, 31,  0,  0},	{ 1,  0,  3,  1,  1,  1,  1, 17, 14},	/* ij */
	{16, 16, 17, 18, 28, 18, 17,  0,  0},	{12,  4,  4,  4,  4,  4, 31,  0,  0},	/* kl */
	{ 0,  0, 30, 21, 21, 21, 21,  0,  0},	{ 0,  0, 30, 17, 17, 17, 17,  0,  0},	/* mn */
	{ 0,  0, 14, 17, 17, 17, 14,  0,  0},	{ 0,  0, 30, 17, 17, 17, 30, 16, 16},	/* op */
	{ 0,  0, 15, 17, 17, 17, 15,  1,  1},	{ 0,  0, 30, 17, 16, 16, 16,  0,  0},	/* qr */
	{ 0,  0, 15, 16, 14,  1, 30,  0,  0},	{ 8,  8, 30,  8,  8,  9,  6,  0,  0},	/* st */
	{ 0,  0, 17, 17, 17, 17, 14,  0,  0},	{ 0,  0, 17, 17, 17, 10,  4,  0,  0},	/* uv */
	{ 0,  0, 17, 21, 21, 21, 10,  0,  0},	{ 0,  0, 17, 10,  4, 10, 17,  0,  0},	/* wx */
	{ 0,  0, 17, 17, 17, 17, 15,  1, 14},	{ 0,  0, 31,  2,  4,  8, 31,  0,  0}	/* yz */
	};

    static char digits[10][9] =
	{
	{14, 17, 17, 17, 17, 17, 14,  0,  0},	{ 2,  6, 10,  2,  2,  2, 31,  0,  0},	/* 01 */
	{14, 17,  2,  4,  8, 16, 31,  0,  0},	{14, 17,  1, 14,  1, 17, 14,  0,  0},	/* 23 */
	{ 2,  6, 10, 31,  2,  2,  2,  0,  0},	{31, 16, 16, 30,  1, 17, 14,  0,  0},	/* 45 */
	{14, 17, 16, 30, 17, 17, 14,  0,  0},	{31,  1,  2,  4,  8, 16, 16,  0,  0},	/* 67 */
	{14, 17, 17, 14, 17, 17, 14,  0,  0},	{14, 17, 17, 15,  1, 17, 14,  0,  0}	/* 89 */
	};
 
    char ch;
 
    for ( ; *str; str++)
	{
	ch = (line >= 0 && line <= 8) ? *str : ' ';
	if (isupper(ch))
	    decode(uppers[ch-'A'][line], c);
	else if (islower(ch))
	    decode(lowers[ch-'a'][line], c);
	else if (isdigit(ch))
	    decode(digits[ch-'0'][line], c);
	else
	    decode(0, c);
	}
    }
