/*
 * $Header: grammar.y,v 1.1 89/03/14 20:59:01 cthuang Exp $
 *
 * yacc grammar for C prototype generator
 * This was derived from the grammar given in Appendix A of
 * "The C Programming Language" by the Kernighan and Ritchie.
 */

/* identifiers that are not reserved words */
%token IDENTIFIER TYPEDEF_NAME

/* storage class */
%token AUTO EXTERN REGISTER STATIC TYPEDEF

/* type specifiers */
%token CHAR DOUBLE FLOAT INT VOID
%token LONG SHORT SIGNED UNSIGNED
%token ENUM STRUCT UNION

/* type qualifiers */
%token CONST VOLATILE

/* paired braces and everything between them: { ... } */
%token BRACES

/* paired square brackets and everything between them: [ ... ]*/
%token BRACKETS

/* three periods */
%token ELLIPSIS

/* equal sign followed by constant expression or stuff between braces */
%token INITIALIZER

%type <decl_spec> decl_specs decl_spec_item
%type <decl_spec> storage_class type_spec type_qualifier
%type <decl_spec> struct_union_spec enum_spec
%type <decl_list> init_decl_list
%type <declarator> init_declarator declarator direct_decl
%type <param_list> param_type_list param_list
%type <parameter> param_decl
%type <param_list> identifier_list ident_list
%type <text> struct_or_union
%type <text> pointer type_qual_list
%type <text> IDENTIFIER TYPEDEF_NAME
%type <text> BRACKETS

%{
#include <stdio.h>
#include "cproto.h"
#include "semantics.h"

/* This variable is TRUE when the parameter declaration part of a function
 * definition is currently being parsed.
 * For example:
 *	int
 *	main (argc, argv)
 *	int argc;		| TRUE when these lines
 *	char **argv;		| are being scanned.
 *	{			 
 *	... 			 
 *      }			 
 */
static boolean func_decl = FALSE;

/* This points to the list of parameters for the current function definition. */
static ParameterList *cur_params;

/* temporary string buffer */
static char buf[MAX_TEXT_LENGTH];

/* Table of typedef names */
SymbolTable *typedef_names;
%}
%%

program		: /* empty */
		| translation_unit
		;

translation_unit : external_decl
		| translation_unit external_decl
		;

external_decl	: declaration
		| function_def
		| error
		;

declaration	: decl_specs ';'
		| decl_specs init_decl_list ';'
		{
		    if (func_decl) {
			set_param_types(cur_params, &$1, &$2);
		    } else {
			output_declarations(&$1, &$2);
		    }
		    free_decl_spec(&$1);
		    free_decl_list(&$2);
		}
		| TYPEDEF decl_specs declarator ';'
		{
		    (void)new_symbol(typedef_names, $3.name);
		    free_decl_spec(&$2);
		    free_declarator(&$3);
		}
		;

function_def	: decl_specs
		  declarator
		{
		    if ($2.func_def == FUNC_DEF_NONE) {
			yyerror("syntax error");
			YYERROR;
		    }
		    func_decl = TRUE;
		    cur_params = &($2.params);
		}
		  opt_decl_list BRACES
		{
		    func_decl = FALSE;

		    output_prototype(&$1, &$2);
		    free_decl_spec(&$1);
		    free_declarator(&$2);
		}
		| declarator
		{
		    if ($1.func_def == FUNC_DEF_NONE) {
			yyerror("syntax error");
			YYERROR;
		    }
		    func_decl = TRUE;
		    cur_params = &($1.params);
		}
		  opt_decl_list BRACES
		{
		    DeclSpec	decl_spec;

		    func_decl = FALSE;

		    new_decl_spec(&decl_spec, "int", DE_EXTERN);
		    output_prototype(&decl_spec, &$1);
		    free_decl_spec(&decl_spec);
		    free_declarator(&$1);
		}
		;

decl_list	: declaration
		| decl_list declaration
		;

decl_specs	: decl_spec_item
		| decl_specs decl_spec_item
		{
		    join_decl_specs(&$$, &$1, &$2);
		}
		;

decl_spec_item	: storage_class
		| type_spec
		| type_qualifier
		;

opt_decl_list	: decl_list
		| /* empty */
		;

storage_class	: AUTO
		{
		    new_decl_spec(&$$, "auto", DE_JUNK);
		}
		| EXTERN
		{
		    new_decl_spec(&$$, "extern", DE_EXTERN);
		}
		| REGISTER
		{
		    new_decl_spec(&$$, "register", DE_JUNK);
		}
		| STATIC
		{
		    new_decl_spec(&$$, "static", DE_STATIC);
		}
		;

type_spec	: CHAR
		{
		    new_decl_spec(&$$, "char", DE_EXTERN);
		}
		| DOUBLE
		{
		    new_decl_spec(&$$, "double", DE_EXTERN);
		}
		| FLOAT
		{
		    new_decl_spec(&$$, "float", DE_EXTERN);
		}
		| INT
		{
		    new_decl_spec(&$$, "int", DE_EXTERN);
		}
		| LONG
		{
		    new_decl_spec(&$$, "long", DE_EXTERN);
		}
		| SHORT
		{
		    new_decl_spec(&$$, "short", DE_EXTERN);
		}
		| SIGNED
		{
		    new_decl_spec(&$$, "signed", DE_EXTERN);
		}
		| UNSIGNED
		{
		    new_decl_spec(&$$, "unsigned", DE_EXTERN);
		}
		| VOID
		{
		    new_decl_spec(&$$, "void", DE_EXTERN);
		}
		| struct_union_spec
		| enum_spec
		| TYPEDEF_NAME
		{
		    new_decl_spec(&$$, $1, DE_EXTERN);
		}
		;

type_qualifier	: CONST
		{
		    new_decl_spec(&$$, "const", DE_EXTERN);
		}
		| VOLATILE
		{
		    new_decl_spec(&$$, "volatile", DE_EXTERN);
		}
		;

struct_union_spec : struct_or_union IDENTIFIER BRACES
		{
		    sprintf(buf, "%s %s {}", $1, $2);
		    new_decl_spec(&$$, buf, DE_EXTERN);
		}
		| struct_or_union BRACES
		{
		    sprintf(buf, "%s {}", $1);
		    new_decl_spec(&$$, buf, DE_EXTERN);
		}
		| struct_or_union IDENTIFIER
		{
		    sprintf(buf, "%s %s", $1, $2);
		    new_decl_spec(&$$, buf, DE_EXTERN);
		}
		;

struct_or_union	: STRUCT
		{
		    strcpy($$, "struct");
		}
		| UNION
		{
		    strcpy($$, "union");
		}
		;

init_decl_list	: init_declarator
		{
		    new_decl_list(&$$, &$1);
		}
		| init_decl_list ',' init_declarator
		{
		    add_decl_list(&$$, &$1, &$3);
		}
		;

init_declarator	: declarator
		| declarator INITIALIZER
		;

enum_spec	: ENUM IDENTIFIER BRACES
		{
		    sprintf(buf, "enum %s {}", $2);
		    new_decl_spec(&$$, buf, DE_EXTERN);
		}
		| ENUM BRACES
		{
		    new_decl_spec(&$$, "enum {}", DE_EXTERN);
		}
		| ENUM IDENTIFIER
		{
		    sprintf(buf, "enum %s", $2);
		    new_decl_spec(&$$, buf, DE_EXTERN);
		}
		;

declarator	: pointer direct_decl
		{
		    sprintf(buf, "%s%s", $1, $2.text);
		    $$ = $2;
		    $$.text = strdup(buf);
		    free($2.text);
		}
		| direct_decl
		;

direct_decl	: IDENTIFIER
		{
		    new_declarator(&$$, $1, $1);
		}
		| '(' declarator ')'
		{
		    sprintf(buf, "(%s)", $2.text);
		    $$ = $2;
		    $$.text = strdup(buf);
		    free($2.text);
		}
		| direct_decl BRACKETS
		{
		    sprintf(buf, "%s%s", $1.text, $2);
		    $$ = $1;
		    $$.text = strdup(buf);
		    free($1.text);
		}
		| direct_decl '(' param_type_list ')'
		{
		    sprintf(buf, "%s(%%s)", $1.text);
		    $$ = $1;
		    $$.func_def = FUNC_DEF_NEW;
		    $$.params = $3;
		    $$.text = strdup(buf);
		    free($1.text);
		}
		| direct_decl '(' identifier_list ')'
		{
		    sprintf(buf, "%s(%%s)", $1.text);
		    $$ = $1;
		    $$.func_def = FUNC_DEF_OLD;
		    $$.params = $3;
		    $$.text = strdup(buf);
		    free($1.text);
		}
		;

pointer		: '*' type_qual_list
		{
		    sprintf($$, "*%s", $2);
		}
		| '*' type_qual_list pointer
		{
		    sprintf($$, "*%s%s", $2, $3);
		}
		;

type_qual_list	: /* empty */
		{
		    strcpy($$, "");
		}
		| type_qual_list type_qualifier
		{
		    sprintf($$, "%s %s ", $1, $2.text);
		    free($2.text);
		}
		;

param_type_list	: VOID
		{
		    Parameter	p;

		    new_parameter(&p, "void", " ", " ");
		    new_param_list(&$$, &p);
		}
		| param_list
		| param_list ',' ELLIPSIS
		{
		    Parameter	p;

		    new_parameter(&p, "...", " ", " ");
		    add_param_list(&$$, &$1, &p);
		}
		;

param_list	: param_decl
		{
		    new_param_list(&$$, &$1);
		}
		| param_list ',' param_decl
		{
		    add_param_list(&$$, &$1, &$3);
		}
		;

param_decl	: decl_specs declarator
		{
		    new_parameter(&$$, $1.text, $2.text, $2.name);
		}
		;

identifier_list	: /* empty */
		{
		    new_ident_list(&$$);
		}
		| ident_list
		;

ident_list	: IDENTIFIER
		{
		    new_ident_list(&$$);
		    add_ident_list(&$$, &$$, $1);
		}
		| ident_list ',' IDENTIFIER
		{
		    add_ident_list(&$$, &$1, $3);
		}
		;

%%
#include "lex.yy.c"

yyerror (msg)
char *msg;
{
    print_error(msg);
}

void
parse_file ()
{
    typedef_names = create_symbol_table();
    (void)yyparse();
}
