/*

setclock.c - get time from the Heathkit GC-1000

Copyright 1990, Nick Sayer
Permission to distribute granted as long as no fee is charged and
those to whom you distribute are also free to do so. All other
rights reserved. No warrantees are extended regarding this software's
correctness or applicability to any specific purpose.

revision history:

v1.4  10/01/90 - minor bug fixes. -d flag now ignores -l. -l 0 disables -l.
v1.3  09/22/90 - added limitation on time "displacement" -l option
v1.2  03/31/90 - added command line option -d (debug)
v1.1  02/01/90 - added configuration section & changed to settimeofday()
v1.0 the epoch - created

BUGS -	adjtime() would be preferable to settimeofday().

				USAGE

% setclock [-d] [-l minutes]

Have cron run this program as often as necessary to maintain system
clock accuracy. If the clock is not set, or if the tenth-of-a-second digit
is dimmed, the system time will not be changed.

If -d is specified, the system will print out what time it is according
to the clock. It will not try to set the system time. Use this to debug
the switch and/or options settings (see below).

If the time is off by more than a certain number of minutes (specified
by -l, default 5), the time will not be changed and an error message
generated instead. The WWV signal has no check-bits and some computers
don't have the best shielding in the world. The result is that despite
the clock's heuristics it can occasionally get the wrong idea of what
time it is. Having the system time suddenly change by a large amount
can cause great confusion, and that confusion is complicated when the
time has to be changed back again. For small numbers of minutes though
this isn't such a big problem. If you know the clock is right and the
system is wrong, you can use "-l 0" to turn this checking off.

			   CONFIGURATION

Your clock should be set to constantly give the time every second or
so. I.e. the handshaking option should be off, and the continuous
option should be on. On my clock, using this configuration, the
time reported is exactly one second slow, so this second is added back
later. To remove this, simply find the line that says
"convert.tm_sec=....." and remove the "+1".

The time sent to the RS-232 reads exactly the same way as it does on
the display, so if your clock displays UT1, so will your unix box.
If the section below is not right, all hell will break loose. :-)

Set DEVICE to the full path of the entry in /dev (or a link to it)
that talks to the clock
*/
#define DEVICE "/dev/clock" /* file for clock device */

/*
Set baud rate for clock (like in sgtty.h)
*/
#define BAUD B2400 /* baud rate for clock */

/*
Set the default minute limit here
*/
#define DEFAULT_MINUS_L 5

/*
If your clock displays AM/PM, define this.
*/
#define AMPM                /* define if clock shows AM/PM */

/*
If you've set the clock up to display local time rather than GMT, then
the clock must also be set properly for your local interpretation of
daylight savings time. If your area has DST, it is NOT a good idea to
run this program less than 24 hours before DST kicks in or turns off.
You may find your clock changing DST status early. Unix will figure out
DST by itself. For no-hassle results, use GMT and undef LOCAL. For
human-readibility of the clock, grin and bear it. :-)
*/
#define LOCAL               /* undef this if your clock displays GMT */

/*
Since the BCD time code doesn't include any indication of what year it
is, heath used internal dipswitches to set the year. Since it's a pain
to get in there on December 31 every year, two options have been
included to get around this. If you define CLOCKYEAR, the clock's idea
of what year it is will be used. If you define YEARFILE, and set it
to a filename, that file is assumed to have two digits and a newline
in it, and those two digits will be used as the year. If you don't
define either, the system's current notion of what year it is will
be retained. If you do it this way, it is NOT a good idea to run
this program close to 23:59:59 on December 31.
*/
/* #define CLOCKYEAR        /* define this to use the clock's year */
/* #define YEARFILE "/usr/local/lib/year" /* define this for file */

/*
For best results, define NICE and set to the priority level desired. -20
is highly recommended. This program doesn't load the CPU much, but
accuracy will suffer if it gets swapped.
*/
#define NICE -20

/*
                        END OF CONFIGURATION
*/

/* spoof 'em for gmtime if necessary */

#ifdef LOCAL
#define LOCALTIME localtime
#define TIMELOCAL timelocal
#else
#define LOCALTIME gmtime
#define TIMELOCAL timegm
#endif

#include <stdio.h>
#include <sgtty.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
extern int errno;               /* perror(3) */
extern char *sys_errlist[];
extern int optind,opterr;       /* getopt(3) */
extern char *optarg;

/* We'll set a timer later. If it goes off, something's messed up. */
bomb()
{

  printf("Timeout. Check clock.\n");
  exit(1);

}

/* A hack. getc() didn't work for some reason. */
char get_c(fildes)
int fildes;
{
  char c;
  read(fildes,&c,1);
  return c;
}

main(argc,argv)
int argc;
char **argv;
{

  int clkfil,dbg_flag=0,usec,hr,timedif=DEFAULT_MINUS_L,errflag=0;
  char a[255],*b,optchar;
  time_t hydrant;

  struct timeval tv;
  struct timezone tz;
  struct tm convert;

  struct sgttyb ttystuff;

#ifdef YEARFILE
  FILE *yf;
  int yf_yr;
#endif

  opterr=0;

  while((optchar=getopt(argc,argv,"dl:"))!=-1)
  {
    switch(optchar)
    {
      case 'd':dbg_flag++; break;
      case 'l':timedif=atoi(optarg); break;
      default:errflag++;
    }
  }
  if (errflag)
  {
    printf("Usage: %s [-d] [-l min]\n",*argv);
    exit(1);
  }

/* if YEARFILE, we'd better read it now so it won't interfere with the
tight timing section later */
#ifdef YEARFILE
  yf=fopen(YEARFILE,"r");
  scanf("%d\n",&yf_yr);
  fclose(yf);
#endif

#ifdef NICE
  nice(NICE);
#endif

  alarm(15);
  signal(SIGALRM,bomb);

  if ((clkfil=open(DEVICE,2))==NULL)
  {
    printf("Error opening %s.\n",DEVICE);
    exit(1);
  }

/* set for baud rate and RAW */
  ttystuff.sg_ispeed=BAUD;
  ttystuff.sg_ospeed=BAUD;
  ttystuff.sg_kill=' ';
  ttystuff.sg_erase=' ';
  ttystuff.sg_flags=RAW;
  if (ioctl(clkfil,TIOCSETP,&ttystuff)==-1)
  {
    printf("Error setting up %s.\n",DEVICE);
    exit(1);
  }
/*
Wait for end of line, then read until another. If seconds>50,
wait for it to go to 0. We have to add 1 to the seconds, and if it's
:59..... Also, if the line's too short, read it again.
*/
  while((get_c(clkfil) & '')!='\r');
  do {
    b=a;
    while((*b=get_c(clkfil) & '')!='\r') b++;
    *b='\0';
  } while (a[6]=='5' || strlen(a)!=23);
/*

     START OF TIME CRITICAL SECTION

From here to end of critical section must be executed as quickly as
possible to (try to) insure accuracy

a[9] is the tenth of a second digit. The clock outputs a ? here if
either the clock is not set, or if it hasn't been updated in 24
hours and therefore isn't sure of itself.
*/
  if (a[9]=='?')
  {
    printf("Clock not set.\n");
    exit(1);
  }

  usec=atoi(a+9)*100;
  hr=atoi(a);

#ifdef AMPM
  if (a[11]=='P') 
  {
    if (hr!=12) hr+=12;
  }
  if (a[11]=='A')
  {
    if (hr==12) hr=0;
  }
#endif

  time(&hydrant);
  convert=*LOCALTIME(&hydrant);
  convert.tm_min=atoi(a+3);
  convert.tm_sec=atoi(a+6)+1;
  convert.tm_hour=hr;
  convert.tm_mon=atoi(a+15)-1;
  convert.tm_mday=atoi(a+18);

#ifdef CLOCKYEAR
  convert.tm_year=atoi(a+21);
#endif

#ifdef YEARFILE
  convert.tm_year=yf_yr;
#endif

  hydrant=TIMELOCAL(&convert);

  if (dbg_flag)
  {
    printf("%s",ctime(&hydrant));
    exit(0);
  }

  if ( (abs(time(0L)-hydrant)>(timedif*60)) && (timedif) )
  {
    printf("Clock shows substantially different time. Check clock.\n");
    exit(1);
  }

/*
Now hydrant is the current ctime (num of secs since epoch)
*/

  gettimeofday(&tv,&tz);
  tv.tv_sec=hydrant;
  tv.tv_usec=usec;
  if (!dbg_flag)
    if (settimeofday(&tv,&tz)!=0)
    {
      printf("settimeofday(): %s\n",sys_errlist[errno]);
      exit(1);
    }

/*

END OF TIME CRITICAL SECTION

*/

exit(0);

}
